package Fractal2D;

import java.util.ArrayList;
import java.util.Iterator;

import Fractal2D.Data.ListMember;
import Fractal2D.Math.TestArithmetics;

import com.jme.image.Texture;
import com.jme.image.Texture2D;
import com.jme.image.Texture.RenderToTextureType;
import com.jme.input.MouseInput;
import com.jme.math.Quaternion;
import com.jme.math.Vector2f;
import com.jme.math.Vector3f;
import com.jme.renderer.ColorRGBA;
import com.jme.renderer.Renderer;
import com.jme.renderer.TextureRenderer;
import com.jme.scene.Spatial;
import com.jme.scene.shape.Quad;
import com.jme.scene.state.GLSLShaderObjectsState;
import com.jme.scene.state.TextureState;
import com.jme.system.DisplaySystem;
import com.jme.system.canvas.SimpleCanvasImpl;

public class FractalImplementor extends SimpleCanvasImpl {
	public final static int NOTHING = 0;
	public final static int CHANGE = 1;
	public final static int BACK = 2;
	public final static int RESET = 3;
	public final static int RESIZE = 4;
	private FractalConfigJPanel oFractalConfigJPanel;
	private TextureRenderer oTextureRenderer;
	private GLSLShaderObjectsState[] aMandelbrot, aJulia;
	private GLSLShaderObjectsState oPositionState;
	// private Vector2f upperRightCorner=new Vector2f(0.8f,1.25f);
	Quad oRenderInitQuad, oQuad, oRenderQuad, oPrecisionQuad, oRenderQuad2,
			oPrecisionQuad2, oPositionQuad;
	Texture2D tDisplay, t1, t2, t3, t4, t5, t6, pos1, pos2;
	// private Vector2
	private boolean pressed = false;
	int switcher = 0;
	int mouseX = 0;
	int mouseY = 0;
	private boolean isMandelbrot = true;
	// private boolean calculating=false;
	// 0=nothing, 1=changed, 2=back, 3=reset, 4=resize
	private int state = 0;
	// private boolean precision=false;
	private Fractal2D oMainFrame;
	private int shaderIndex = 0;
	private float iterationsPerStep = 250;
	private float iterationsPerStep2 = 20;
	private float tempIterations = 0;
	private float maxIterations = 0;
	private ArrayList<Spatial> oSpatList1, oSpatList2, oSpatList3, oSpatList4,
			oSpatList5, oSpatList6;
	private ArrayList<Texture> oTextList1, oTextList2, oTextList3, oTextList4,
			oTextList5;
	TextureState oState, oRenderState, oRenderState2, oRenderPrecisionState,
			oRenderPrecisionState2, oRenderInitState;
	private boolean ping = true;

	protected FractalImplementor(int width, int height,
			FractalConfigJPanel oFractalConfigJPanel, Fractal2D oMainFrame) {
		super(width, height);
		this.oFractalConfigJPanel = oFractalConfigJPanel;
		this.oMainFrame = oMainFrame;
	}

	/**
	 * loads all shaders
	 */
	private void loadShaders() {
		aMandelbrot = new GLSLShaderObjectsState[6];
		aMandelbrot[0] = renderer.createGLSLShaderObjectsState();
		aMandelbrot[0].load(getClass().getClassLoader().getResource(
				"Fractal2D/Shader/initFractal1.vert"), getClass()
				.getClassLoader().getResource(
						"Fractal2D/Shader/initMandelbrot1.frag"));
		aMandelbrot[0].setEnabled(true);
		aJulia = new GLSLShaderObjectsState[6];
		aJulia[0] = renderer.createGLSLShaderObjectsState();
		aJulia[0].load(getClass().getClassLoader().getResource(
				"Fractal2D/Shader/initFractal1.vert"), getClass()
				.getClassLoader().getResource(
						"Fractal2D/Shader/initJulia1.frag"));
		aJulia[0].setEnabled(true);
		// 1
		aMandelbrot[1] = renderer.createGLSLShaderObjectsState();
		aMandelbrot[1].load(getClass().getClassLoader().getResource(
				"Fractal2D/Shader/initFractal1.vert"), getClass()
				.getClassLoader().getResource(
						"Fractal2D/Shader/initMandelbrot2.frag"));
		aMandelbrot[1].setEnabled(true);
		aJulia[1] = renderer.createGLSLShaderObjectsState();
		aJulia[1].load(getClass().getClassLoader().getResource(
				"Fractal2D/Shader/initFractal1.vert"), getClass()
				.getClassLoader().getResource(
						"Fractal2D/Shader/initJulia2.frag"));
		aJulia[1].setEnabled(true);

		// 2
		aMandelbrot[2] = renderer.createGLSLShaderObjectsState();
		aMandelbrot[2].load(getClass().getClassLoader().getResource(
				"Fractal2D/Shader/MandelbrotStep1.vert"), getClass()
				.getClassLoader().getResource(
						"Fractal2D/Shader/MandelbrotStep1.frag"));
		aMandelbrot[2].setEnabled(true);

		aJulia[2] = renderer.createGLSLShaderObjectsState();
		aJulia[2].load(getClass().getClassLoader().getResource(
				"Fractal2D/Shader/JuliaStep1.vert"), getClass()
				.getClassLoader().getResource(
						"Fractal2D/Shader/JuliaStep1.frag"));
		aJulia[2].setEnabled(true);
		// 3
		aMandelbrot[3] = renderer.createGLSLShaderObjectsState();
		aMandelbrot[3].load(getClass().getClassLoader().getResource(
				"Fractal2D/Shader/getValues.vert"), getClass().getClassLoader()
				.getResource("Fractal2D/Shader/initMandelbrot3.frag"));
		aMandelbrot[3].setEnabled(true);
		aJulia[3] = renderer.createGLSLShaderObjectsState();
		aJulia[3].load(getClass().getClassLoader().getResource(
				"Fractal2D/Shader/getValues.vert"), getClass().getClassLoader()
				.getResource("Fractal2D/Shader/initJulia3.frag"));
		aJulia[3].setEnabled(true);
		// 4
		aMandelbrot[4] = renderer.createGLSLShaderObjectsState();
		aMandelbrot[4].load(getClass().getClassLoader().getResource(
				"Fractal2D/Shader/getValues.vert"), getClass().getClassLoader()
				.getResource("Fractal2D/Shader/initMandelbrot4.frag"));
		aMandelbrot[4].setEnabled(true);
		aJulia[4] = renderer.createGLSLShaderObjectsState();
		aJulia[4].load(getClass().getClassLoader().getResource(
				"Fractal2D/Shader/getValues.vert"), getClass().getClassLoader()
				.getResource("Fractal2D/Shader/initJulia4.frag"));
		aJulia[4].setEnabled(true);
		// 5
		aMandelbrot[5] = renderer.createGLSLShaderObjectsState();
		aMandelbrot[5].load(getClass().getClassLoader().getResource(
				"Fractal2D/Shader/getValues.vert"), getClass().getClassLoader()
				.getResource("Fractal2D/Shader/MandelbrotStep2.frag"));
		aMandelbrot[5].setEnabled(true);
		aJulia[5] = renderer.createGLSLShaderObjectsState();
		aJulia[5].load(getClass().getClassLoader().getResource(
				"Fractal2D/Shader/JuliaStep2.vert"), getClass()
				.getClassLoader().getResource(
						"Fractal2D/Shader/JuliaStep2.frag"));
		aJulia[5].setEnabled(true);
		oPositionState = renderer.createGLSLShaderObjectsState();
		oPositionState.load(getClass().getClassLoader().getResource(
				"Fractal2D/Shader/initFractal2.vert"), getClass()
				.getClassLoader().getResource(
						"Fractal2D/Shader/convertValues.frag"));
		oPositionState.setEnabled(true);
	}

	/**
	 * creates the textures for rendering
	 */
	public void simpleSetup() {
		try {
			// DisplaySystem.getDisplaySystem().setVSyncEnabled(true);
			renderer.setBackgroundColor(ColorRGBA.white);
			if (!GLSLShaderObjectsState.isSupported()) {
				this.oMainFrame.stopIt("Your system does not support GLSL.");
				return;
			}
			this.loadShaders();
			oTextureRenderer = DisplaySystem.getDisplaySystem()
					.createTextureRenderer(width, height,
							TextureRenderer.Target.Texture2D);
			oTextureRenderer.setMultipleTargets(true);
			t1 = new Texture2D();
			t1.setRenderToTextureType(RenderToTextureType.RGBA32F);
			t2 = new Texture2D();
			t2.setRenderToTextureType(RenderToTextureType.RGBA32F);
			t3 = new Texture2D();
			t3.setRenderToTextureType(RenderToTextureType.RGBA32F);
			t4 = new Texture2D();
			t4.setRenderToTextureType(RenderToTextureType.RGBA32F);
			t5 = new Texture2D();
			t5.setRenderToTextureType(RenderToTextureType.RGBA32F);
			t6 = new Texture2D();
			t6.setRenderToTextureType(RenderToTextureType.RGBA32F);
			tDisplay = new Texture2D();
			tDisplay.setRenderToTextureType(RenderToTextureType.RGBA32F);
			pos1 = new Texture2D();
			pos1.setRenderToTextureType(RenderToTextureType.RGBA32F);
			pos2 = new Texture2D();
			pos2.setRenderToTextureType(RenderToTextureType.RGBA32F);
			oTextureRenderer.setupTexture(tDisplay);
			oTextureRenderer.setupTexture(t1);
			oTextureRenderer.setupTexture(t2);
			oTextureRenderer.setupTexture(t3);
			oTextureRenderer.setupTexture(t4);
			oTextureRenderer.setupTexture(t5);
			oTextureRenderer.setupTexture(t6);
			oTextureRenderer.setupTexture(pos1);
			oTextureRenderer.setupTexture(pos2);
			oState = renderer.createTextureState();
			oState.setEnabled(true);
			oState.setTexture(tDisplay, 0);

			oRenderState = renderer.createTextureState();
			oRenderState.setTexture(t1, 0);
			oRenderState.setEnabled(true);

			oRenderState2 = renderer.createTextureState();
			oRenderState2.setTexture(t2, 0);
			oRenderState2.setEnabled(true);

			oRenderPrecisionState = renderer.createTextureState();
			oRenderPrecisionState.setTexture(t1, 0);
			oRenderPrecisionState.setTexture(t2, 1);
			oRenderPrecisionState.setTexture(t3, 2);
			oRenderPrecisionState.setTexture(pos1, 3);
			oRenderPrecisionState.setTexture(pos2, 4);
			oRenderPrecisionState.setEnabled(true);

			oRenderPrecisionState2 = renderer.createTextureState();
			oRenderPrecisionState2.setTexture(t4, 0);
			oRenderPrecisionState2.setTexture(t5, 1);
			oRenderPrecisionState2.setTexture(t6, 2);
			oRenderPrecisionState2.setTexture(pos1, 3);
			oRenderPrecisionState2.setTexture(pos2, 4);
			oRenderPrecisionState2.setEnabled(true);

			// display quad
			oQuad = new Quad("myQuad", width, height);
			oQuad.setLocalTranslation(new Vector3f(((float) width) / 2,
					((float) height) / 2, 0));
			oQuad.setRenderQueueMode(Renderer.QUEUE_ORTHO);
			oQuad.setRenderState(oState);
			// oQuad.setRenderState(so);
			rootNode.attachChild(oQuad);
			rootNode.updateRenderState();

			oRenderInitQuad = new Quad("RenderInitMandelbrotQuad", width,
					height);
			oRenderInitQuad.setLocalTranslation(new Vector3f(
					((float) width) / 2, ((float) height) / 2, 0));
			oRenderInitQuad.setRenderQueueMode(Renderer.QUEUE_ORTHO);
			oRenderInitState = renderer.createTextureState();
			oRenderInitState.setTexture(pos1, 0);
			oRenderInitState.setTexture(pos2, 1);
			oRenderInitState.setEnabled(true);
			oRenderInitQuad.setRenderState(oRenderInitState);
			// oRenderInitQuad.setRenderState(oState);
			this.isMandelbrot = this.oFractalConfigJPanel.isMandelbrot();
			if (this.isMandelbrot)
				oRenderInitQuad.setRenderState(aMandelbrot[shaderIndex]);
			else
				oRenderInitQuad.setRenderState(aJulia[shaderIndex]);
			oRenderInitQuad.updateRenderState();

			oRenderQuad = new Quad("RenderQuad", width, height);
			oRenderQuad.setLocalTranslation(new Vector3f(((float) width) / 2,
					((float) height) / 2, 0));
			oRenderQuad.setRenderQueueMode(Renderer.QUEUE_ORTHO);
			oRenderQuad.setRenderState(oRenderState);
			this.isMandelbrot = this.oFractalConfigJPanel.isMandelbrot();
			if (this.isMandelbrot)
				oRenderQuad.setRenderState(aMandelbrot[2]);
			else
				oRenderQuad.setRenderState(aJulia[2]);
			oRenderQuad.updateRenderState();

			oRenderQuad2 = new Quad("RenderQuad2", width, height);
			oRenderQuad2.setLocalTranslation(new Vector3f(((float) width) / 2,
					((float) height) / 2, 0));
			oRenderQuad2.setRenderQueueMode(Renderer.QUEUE_ORTHO);
			oRenderQuad2.setRenderState(oRenderState2);
			if (this.isMandelbrot)
				oRenderQuad2.setRenderState(aMandelbrot[2]);
			else
				oRenderQuad2.setRenderState(aJulia[2]);
			oRenderQuad2.updateRenderState();

			oPrecisionQuad = new Quad("PrecisionQuad", width, height);
			oPrecisionQuad.setLocalTranslation(new Vector3f(
					((float) width) / 2, ((float) height) / 2, 0));
			oPrecisionQuad.setRenderQueueMode(Renderer.QUEUE_ORTHO);
			oPrecisionQuad.setRenderState(oRenderPrecisionState);
			if (this.isMandelbrot)
				oPrecisionQuad.setRenderState(aMandelbrot[5]);
			else
				oPrecisionQuad.setRenderState(aJulia[5]);
			oPrecisionQuad.updateRenderState();

			oPrecisionQuad2 = new Quad("PrecisionQuad2", width, height);
			oPrecisionQuad2.setLocalTranslation(new Vector3f(
					((float) width) / 2, ((float) height) / 2, 0));
			oPrecisionQuad2.setRenderQueueMode(Renderer.QUEUE_ORTHO);
			oPrecisionQuad2.setRenderState(oRenderPrecisionState2);

			if (this.isMandelbrot)
				oPrecisionQuad2.setRenderState(aMandelbrot[5]);
			else
				oPrecisionQuad2.setRenderState(aJulia[5]);
			oPrecisionQuad2.updateRenderState();

			oPositionQuad = new Quad("PositionQuad", width, height);
			oPositionQuad.setLocalTranslation(new Vector3f(((float) width) / 2,
					((float) height) / 2, 0));
			oPositionQuad.setRenderQueueMode(Renderer.QUEUE_ORTHO);
			oPositionQuad.setRenderState(oPositionState);
			oPositionQuad.updateRenderState();
			oSpatList1 = new ArrayList<Spatial>();
			oTextList1 = new ArrayList<Texture>();
			oSpatList1.add(oRenderInitQuad);
			oTextList1.add(tDisplay);
			oTextList1.add(t2);
			oSpatList2 = new ArrayList<Spatial>();
			oSpatList2.add(oRenderQuad);
			oTextList2 = new ArrayList<Texture>();
			oTextList2.add(tDisplay);
			oTextList2.add(t4);
			oTextList2.add(t5);
			oTextList2.add(t6);
			oSpatList3 = new ArrayList<Spatial>();
			oSpatList3.add(oPrecisionQuad);
			oSpatList4 = new ArrayList<Spatial>();
			oSpatList4.add(oPositionQuad);
			oTextList3 = new ArrayList<Texture>();
			oTextList3.add(pos1);
			oTextList3.add(pos2);
			oTextList4 = new ArrayList<Texture>();
			oTextList4.add(tDisplay);
			oTextList4.add(t1);
			oTextList5 = new ArrayList<Texture>();
			oTextList5.add(tDisplay);
			oTextList5.add(t1);
			oTextList5.add(t2);
			oTextList5.add(t3);
			oSpatList5 = new ArrayList<Spatial>();
			oSpatList5.add(oRenderQuad2);
			oSpatList6 = new ArrayList<Spatial>();
			oSpatList6.add(oPrecisionQuad2);
			calculate();
			this.renderer.checkCardError();
		} catch (Exception ex) {
			String exString = ex.toString() + " at:\n";
			StackTraceElement[] elements = ex.getStackTrace();
			for (int i = 0; i < elements.length; i++, exString += "\n") {
				exString += elements[i].toString();
			}
			this.oMainFrame.stopIt(exString);
			return;
		}
	}

	/**
	 * fictive formula to adjust the iterations
	 * 
	 * @return
	 */
	private float adjustIterations() {
		double allZooms = 1;
		ArrayList<ListMember> oZoomList = this.isMandelbrot ? this.oFractalConfigJPanel.oZoomMandelbrot
				: this.oFractalConfigJPanel.oZoomJulia;
		for (Iterator<ListMember> oIterator = oZoomList.iterator(); oIterator
				.hasNext();) {
			allZooms *= oIterator.next().zoom;
		}
		float iterations = (float) (Math.rint(63 * Math.pow(1.33, Math
				.log10(allZooms))));
		if (iterations > 100000)
			iterations = 100000;
		return iterations;
	}

	/**
	 * single precision corners
	 * 
	 * @return
	 */
	private Vector2f[] getCorners() {
		Vector2f lowerLeftCorner;
		Vector2f upperRightCorner;
		ArrayList<ListMember> oZoomList;
		if (this.isMandelbrot) {
			lowerLeftCorner = new Vector2f(-2.2f, -1.25f);
			upperRightCorner = new Vector2f(0.8f, 1.25f);
			oZoomList = this.oFractalConfigJPanel.oZoomMandelbrot;

		} else {
			lowerLeftCorner = new Vector2f(-2f, -2f);
			upperRightCorner = new Vector2f(2f, 2f);
			oZoomList = this.oFractalConfigJPanel.oZoomJulia;
		}
		for (Iterator<ListMember> oIterator = oZoomList.iterator(); oIterator
				.hasNext();) {
			ListMember oListMember = oIterator.next();
			float dx = upperRightCorner.x - lowerLeftCorner.x;
			float dy = upperRightCorner.y - lowerLeftCorner.y;
			float centerX = lowerLeftCorner.x + dx * (float) oListMember.x
					/ (float) oListMember.width;
			float centerY = lowerLeftCorner.y + dy * (float) oListMember.y
					/ (float) oListMember.height;
			lowerLeftCorner.x = (float) (centerX - dx / 2 / oListMember.zoom);
			upperRightCorner.x = (float) (centerX + dx / 2 / oListMember.zoom);
			lowerLeftCorner.y = (float) (centerY - dy / 2 / oListMember.zoom);
			upperRightCorner.y = (float) (centerY + dy / 2 / oListMember.zoom);
			// if not precise enough, return null
			if (2.0f + (float) lowerLeftCorner.x
					+ ((float) upperRightCorner.x - (float) lowerLeftCorner.x)
					/ 2.0f / (float) width <= (float) lowerLeftCorner.x + 2.0f
					|| 2.0f
							+ (float) lowerLeftCorner.y
							+ ((float) upperRightCorner.y - (float) lowerLeftCorner.y)
							/ 2.0f / (float) height <= (float) lowerLeftCorner.y + 2.0f) {
				return null;
			}
		}
		return new Vector2f[] { lowerLeftCorner, upperRightCorner };
	}

	/**
	 * calculates the corners with high precision
	 * 
	 * @return
	 */
	private TestArithmetics[] getAdvancedCorners() {
		TestArithmetics lowerLeftCornerX;
		TestArithmetics lowerLeftCornerY;
		TestArithmetics upperRightCornerX;
		TestArithmetics upperRightCornerY;
		ArrayList<ListMember> oZoomList;
		if (this.isMandelbrot) {
			lowerLeftCornerX = new TestArithmetics(-2.2, 8);
			lowerLeftCornerY = new TestArithmetics(-1.25, 8);
			upperRightCornerX = new TestArithmetics(0.8, 8);
			upperRightCornerY = new TestArithmetics(1.25, 8);
			oZoomList = this.oFractalConfigJPanel.oZoomMandelbrot;
		} else {
			lowerLeftCornerX = new TestArithmetics(-2, 8);
			lowerLeftCornerY = new TestArithmetics(-2, 8);
			upperRightCornerX = new TestArithmetics(2, 8);
			upperRightCornerY = new TestArithmetics(2, 8);
			oZoomList = this.oFractalConfigJPanel.oZoomJulia;
		}
		for (Iterator<ListMember> oIterator = oZoomList.iterator(); oIterator
				.hasNext();) {
			ListMember oListMember = oIterator.next();
			TestArithmetics dx = upperRightCornerX.clone();
			// dx
			dx.add(lowerLeftCornerX.chgSign());
			TestArithmetics dy = upperRightCornerY.clone();
			// dy
			dy.add(lowerLeftCornerY.chgSign());
			// centerX
			TestArithmetics oListMemberX = new TestArithmetics(oListMember.x, 8);
			oListMemberX.mult(dx);
			oListMemberX.div(oListMember.width);
			TestArithmetics centerX = lowerLeftCornerX.clone();
			centerX.add(oListMemberX);
			// centerY
			TestArithmetics oListMemberY = new TestArithmetics(oListMember.y, 8);
			oListMemberY.mult(dy);
			oListMemberY.div(oListMember.height);
			TestArithmetics centerY = lowerLeftCornerY.clone();
			centerY.add(oListMemberY);
			// corners
			// x
			dx.div(2);
			dx.div(oListMember.zoom);
			lowerLeftCornerX = centerX.clone();
			lowerLeftCornerX.add(dx.chgSign());
			upperRightCornerX = centerX.clone();
			upperRightCornerX.add(dx);
			// y
			dy.div(2);
			dy.div(oListMember.zoom);
			lowerLeftCornerY = centerY.clone();
			lowerLeftCornerY.add(dy.chgSign());
			upperRightCornerY = centerY.clone();
			upperRightCornerY.add(dy);
		}

		return new TestArithmetics[] { lowerLeftCornerX, lowerLeftCornerY,
				upperRightCornerX, upperRightCornerY };
	}

	public void setState(int state) {
		this.state = state;
	}

	public float getAdjustedIterations() {
		return adjustIterations();
	}

	/**
	 * calculates the boundaries for the fractal and gives them to the shader;
	 * after that, the fractal will be rendered
	 */
	private void calculate() {
		try {
			if (this.oFractalConfigJPanel.adjustIterations()) {
				this.maxIterations = adjustIterations();
				this.oFractalConfigJPanel
						.setIterations((int) this.maxIterations);
			} else
				this.maxIterations = (float) this.oFractalConfigJPanel
						.getIterations();

			Vector2f[] aVector = getCorners();

			// if(this.tempIterations<this.maxIterations){
			this.oFractalConfigJPanel.showProgress(true);
			this.oFractalConfigJPanel.setProgressValue(0);
			// }else this.oFractalConfigJPanel.showProgress(false);
			// this.maxIterations = 500;
			// aVector = null;
			if (aVector == null) {
				if (this.shaderIndex <= 2) {
					oMainFrame.setTitle("Fractal2D - Precision mode");
				}
				if (this.maxIterations > this.iterationsPerStep2) {
					this.tempIterations = this.iterationsPerStep2;
					this.shaderIndex = 4;
				} else {
					this.tempIterations = this.maxIterations;
					this.shaderIndex = 3;
				}
				TestArithmetics[] aCorners = getAdvancedCorners();
				oPositionState.setUniform("lowerLeftCornerX1", new Quaternion(
						aCorners[0].buffer[0], aCorners[0].buffer[1],
						aCorners[0].buffer[2], aCorners[0].buffer[3]));
				oPositionState.setUniform("lowerLeftCornerX2", new Quaternion(
						aCorners[0].buffer[4], aCorners[0].buffer[5],
						aCorners[0].buffer[6], aCorners[0].buffer[7]));
				oPositionState.setUniform("lowerLeftCornerY1", new Quaternion(
						aCorners[1].buffer[0], aCorners[1].buffer[1],
						aCorners[1].buffer[2], aCorners[1].buffer[3]));
				oPositionState.setUniform("lowerLeftCornerY2", new Quaternion(
						aCorners[1].buffer[4], aCorners[1].buffer[5],
						aCorners[1].buffer[6], aCorners[1].buffer[7]));
				TestArithmetics deltaX = aCorners[2].clone();
				deltaX.add(aCorners[0].chgSign());
				TestArithmetics deltaY = aCorners[3].clone();
				deltaY.add(aCorners[1].chgSign());
				deltaX.div(2);
				deltaY.div(2);
				oPositionState.setUniform("deltaVectorX1", new Quaternion(
						deltaX.buffer[0], deltaX.buffer[1], deltaX.buffer[2],
						deltaX.buffer[3]));
				oPositionState.setUniform("deltaVectorX2", new Quaternion(
						deltaX.buffer[4], deltaX.buffer[5], deltaX.buffer[6],
						deltaX.buffer[7]));
				oPositionState.setUniform("deltaVectorY1", new Quaternion(
						deltaY.buffer[0], deltaY.buffer[1], deltaY.buffer[2],
						deltaY.buffer[3]));
				oPositionState.setUniform("deltaVectorY2", new Quaternion(
						deltaY.buffer[4], deltaY.buffer[5], deltaY.buffer[6],
						deltaY.buffer[7]));
				if (this.isMandelbrot) {

					this.aMandelbrot[3].setUniform("inputTexture1", 0);
					this.aMandelbrot[3].setUniform("inputTexture2", 1);
					aMandelbrot[3].setUniform("maxIter", this.tempIterations);
					this.aMandelbrot[4].setUniform("inputTexture1", 0);
					this.aMandelbrot[4].setUniform("inputTexture2", 1);
					aMandelbrot[4].setUniform("maxIter", this.tempIterations);
					// System.out.println(this.tempIterations);
					this.oRenderInitQuad
							.setRenderState(this.aMandelbrot[shaderIndex]);
					this.oRenderInitQuad.updateRenderState();
				} else {
					aJulia[3].setUniform("maxIter", this.tempIterations);
					// if(this.maxIterations>this.iterationsPerStep)aJulia[
					// shaderIndex
					// ].setEnabled(false);
					double[] cValue = this.oFractalConfigJPanel.getC();
					TestArithmetics cX = new TestArithmetics(cValue[0], 8);
					TestArithmetics cY = new TestArithmetics(cValue[1], 8);
					aJulia[3].setUniform("cX1", new Quaternion(cX.buffer[0],
							cX.buffer[1], cX.buffer[2], cX.buffer[3]));
					aJulia[3].setUniform("cX2", new Quaternion(cX.buffer[4],
							cX.buffer[5], cX.buffer[6], cX.buffer[7]));
					aJulia[3].setUniform("cY1", new Quaternion(cY.buffer[0],
							cY.buffer[1], cY.buffer[2], cY.buffer[3]));
					aJulia[3].setUniform("cY2", new Quaternion(cY.buffer[4],
							cY.buffer[5], cY.buffer[6], cY.buffer[7]));
					this.aJulia[3].setUniform("inputTexture1", 0);
					this.aJulia[3].setUniform("inputTexture2", 1);
					aJulia[4].setUniform("maxIter", this.tempIterations);
					aJulia[4].setUniform("cX1", new Quaternion(cX.buffer[0],
							cX.buffer[1], cX.buffer[2], cX.buffer[3]));
					aJulia[4].setUniform("cX2", new Quaternion(cX.buffer[4],
							cX.buffer[5], cX.buffer[6], cX.buffer[7]));
					aJulia[4].setUniform("cY1", new Quaternion(cY.buffer[0],
							cY.buffer[1], cY.buffer[2], cY.buffer[3]));
					aJulia[4].setUniform("cY2", new Quaternion(cY.buffer[4],
							cY.buffer[5], cY.buffer[6], cY.buffer[7]));
					aJulia[4].setUniform("inputTexture1", 0);
					aJulia[4].setUniform("inputTexture2", 1);
					aJulia[5].setUniform("cX1", new Quaternion(cX.buffer[0],
							cX.buffer[1], cX.buffer[2], cX.buffer[3]));
					aJulia[5].setUniform("cX2", new Quaternion(cX.buffer[4],
							cX.buffer[5], cX.buffer[6], cX.buffer[7]));
					aJulia[5].setUniform("cY1", new Quaternion(cY.buffer[0],
							cY.buffer[1], cY.buffer[2], cY.buffer[3]));
					aJulia[5].setUniform("cY2", new Quaternion(cY.buffer[4],
							cY.buffer[5], cY.buffer[6], cY.buffer[7]));
					this.oRenderInitQuad
							.setRenderState(this.aJulia[shaderIndex]);
					this.oRenderInitQuad.updateRenderState();
				}
				oTextureRenderer.render(oSpatList4, oTextList3, true);
				if (this.shaderIndex == 3) {
					oTextureRenderer.render(oRenderInitQuad, tDisplay, false);
				} else {
					oTextureRenderer.render(oSpatList1, oTextList2, false);

				}

			} else {
				if (this.shaderIndex > 2) {
					oMainFrame.setTitle("Fractal2D - Speed mode");
				}
				if (this.maxIterations > this.iterationsPerStep) {
					this.tempIterations = this.iterationsPerStep;
					this.shaderIndex = 1;
				} else {
					this.tempIterations = this.maxIterations;
					this.shaderIndex = 0;
				}

				if (this.isMandelbrot) {
					aMandelbrot[0].setUniform("lowerLeftCorner", aVector[0]);
					aMandelbrot[0].setUniform("deltaVector", new Vector2f(
							(aVector[1].x - aVector[0].x) / 2,
							(aVector[1].y - aVector[0].y) / 2));
					aMandelbrot[0].setUniform("maxIter", this.tempIterations);

					aMandelbrot[1].setUniform("lowerLeftCorner", aVector[0]);
					aMandelbrot[1].setUniform("deltaVector", new Vector2f(
							(aVector[1].x - aVector[0].x) / 2,
							(aVector[1].y - aVector[0].y) / 2));
					aMandelbrot[1].setUniform("maxIter", this.tempIterations);
					aMandelbrot[2].setUniform("lowerLeftCorner", aVector[0]);
					aMandelbrot[2].setUniform("deltaVector", new Vector2f(
							(aVector[1].x - aVector[0].x) / 2,
							(aVector[1].y - aVector[0].y) / 2));
					this.oRenderInitQuad
							.setRenderState(this.aMandelbrot[shaderIndex]);
					this.oRenderInitQuad.updateRenderState();

				} else {
					aJulia[0].setUniform("deltaVector", new Vector2f(
							(aVector[1].x - aVector[0].x) / 2,
							(aVector[1].y - aVector[0].y) / 2));
					aJulia[0].setUniform("lowerLeftCorner", aVector[0]);
					aJulia[0].setUniform("maxIter", this.tempIterations);
					double[] cValue = this.oFractalConfigJPanel.getC();
					aJulia[0].setUniform("c", new Vector2f((float) cValue[0],
							(float) cValue[1]));
					aJulia[1].setUniform("deltaVector", new Vector2f(
							(aVector[1].x - aVector[0].x) / 2,
							(aVector[1].y - aVector[0].y) / 2));
					aJulia[1].setUniform("lowerLeftCorner", aVector[0]);
					aJulia[1].setUniform("maxIter", this.tempIterations);
					aJulia[1].setUniform("c", new Vector2f((float) cValue[0],
							(float) cValue[1]));
					aJulia[2].setUniform("c", new Vector2f((float) cValue[0],
							(float) cValue[1]));
					this.oRenderInitQuad
							.setRenderState(this.aJulia[shaderIndex]);
					this.oRenderInitQuad.updateRenderState();
				}

				if (this.shaderIndex == 0) {
					oTextureRenderer.render(oRenderInitQuad, tDisplay, true);
				} else {
					oTextureRenderer.render(oSpatList1, oTextList1, true);

				}

			}
			this.renderer.finish();
			this.renderer.checkCardError();

			// System.out.println(this.shaderIndex);

			double progressValue = (this.tempIterations <= this.maxIterations) ? this.tempIterations
					/ this.maxIterations * 100
					: 100;
			this.oFractalConfigJPanel.setProgressValue(progressValue);

		} catch (Exception ex) {
			// System.out.println(ex);
			String exString = ex.toString() + " at:\n";
			StackTraceElement[] elements = ex.getStackTrace();
			for (int i = 0; i < elements.length; i++, exString += "\n") {
				exString += elements[i].toString();
			}
			this.oMainFrame.stopIt(exString);
			return;
		}
		// System.out.println(this.shaderIndex);
		ping = false;
	}

	public void simpleUpdate() {
		try {
			ArrayList<ListMember> oZoomList = this.isMandelbrot ? this.oFractalConfigJPanel.oZoomMandelbrot
					: this.oFractalConfigJPanel.oZoomJulia;
			if (MouseInput.get().isButtonDown(0)) {
				pressed = true;
				mouseX = MouseInput.get().getXAbsolute();
				mouseY = MouseInput.get().getYAbsolute();
			} else if (pressed) {
				pressed = false;

				if (mouseX >= 0 && mouseX < width && mouseY >= 0
						&& mouseY < height) {

					double zoom = this.oFractalConfigJPanel.getZoom();
					ListMember oListMember = new ListMember(mouseX, height
							- mouseY - 1, width, height, zoom);

					oZoomList.add(oListMember);
					calculate();

				}
			} else {
				switch (state) {
				case NOTHING: {
					if ((this.shaderIndex == 1 || this.shaderIndex == 4)
							&& this.tempIterations < this.maxIterations) {
						if (this.shaderIndex == 1) {
							if (this.isMandelbrot) {
								if (this.tempIterations
										+ this.iterationsPerStep < this.maxIterations) {
									this.tempIterations += this.iterationsPerStep;
								} else {
									this.tempIterations = this.maxIterations;
								}
								this.aMandelbrot[2].setUniform("maxIter",
										this.tempIterations);
								this.aMandelbrot[2].setUniform("inputTexture",
										0);
							} else {
								if (this.tempIterations
										+ this.iterationsPerStep < this.maxIterations) {
									this.tempIterations += this.iterationsPerStep;
								} else {
									this.tempIterations = this.maxIterations;
								}
								this.aJulia[2].setUniform("maxIter",
										this.tempIterations);
								this.aJulia[2].setUniform("inputTexture", 0);
							}
							if (ping) {
								this.oTextureRenderer.render(this.oSpatList2,
										this.oTextList1, false);
							} else {
								this.oTextureRenderer.render(this.oSpatList5,
										this.oTextList4, false);
							}
							this.renderer.finish();
							this.oFractalConfigJPanel
									.setProgressValue(this.tempIterations
											/ this.maxIterations * 100);
						} else {
							if (this.isMandelbrot) {
								if (this.tempIterations
										+ this.iterationsPerStep2 < this.maxIterations) {
									this.tempIterations += this.iterationsPerStep2;
								} else {
									this.tempIterations = this.maxIterations;
								}
								this.aMandelbrot[5].setUniform("maxIter",
										this.tempIterations);
								this.aMandelbrot[5].setUniform("inputTexture1",
										0);
								this.aMandelbrot[5].setUniform("inputTexture2",
										1);
								this.aMandelbrot[5].setUniform("inputTexture3",
										2);
								this.aMandelbrot[5].setUniform("inputTexture4",
										3);
								this.aMandelbrot[5].setUniform("inputTexture5",
										4);
							} else {
								if (this.tempIterations
										+ this.iterationsPerStep2 < this.maxIterations) {
									this.tempIterations += this.iterationsPerStep2;
								} else {
									this.tempIterations = this.maxIterations;
								}
								this.aJulia[5].setUniform("maxIter",
										this.tempIterations);
								this.aJulia[5].setUniform("inputTexture1", 0);
								this.aJulia[5].setUniform("inputTexture2", 1);
								this.aJulia[5].setUniform("inputTexture3", 2);

							}
							if (ping) {
								this.oTextureRenderer.render(this.oSpatList3,
										this.oTextList2, false);

							} else {
								this.oTextureRenderer.render(this.oSpatList6,
										this.oTextList5, false);
							}
							ping = !ping;
							this.renderer.finish();
							this.renderer.checkCardError();
							
							this.oFractalConfigJPanel
									.setProgressValue(this.tempIterations
											/ this.maxIterations * 100);

						}
					}

					break;
				}
				case CHANGE: {

					if (this.isMandelbrot != this.oFractalConfigJPanel
							.isMandelbrot()) {
						this.isMandelbrot = this.oFractalConfigJPanel
								.isMandelbrot();
						if (this.isMandelbrot) {
							this.oRenderInitQuad
									.setRenderState(this.aMandelbrot[shaderIndex]);
							this.oRenderQuad
									.setRenderState(this.aMandelbrot[2]);
							this.oRenderQuad2
									.setRenderState(this.aMandelbrot[2]);
							this.oPrecisionQuad
									.setRenderState(this.aMandelbrot[5]);
							this.oPrecisionQuad2
									.setRenderState(this.aMandelbrot[5]);
						} else {
							this.oRenderInitQuad
									.setRenderState(this.aJulia[shaderIndex]);
							this.oRenderQuad.setRenderState(this.aJulia[2]);
							this.oPrecisionQuad.setRenderState(this.aJulia[5]);
							this.oRenderQuad2.setRenderState(this.aJulia[2]);
							this.oPrecisionQuad2.setRenderState(this.aJulia[5]);
						}
						this.oRenderInitQuad.updateRenderState();
						this.oRenderQuad.updateRenderState();
						this.oPrecisionQuad.updateRenderState();
						this.oRenderQuad2.updateRenderState();
						this.oPrecisionQuad2.updateRenderState();
					}

					calculate();
					break;
				}
				case BACK: {
					if (oZoomList.size() > 0)
						oZoomList.remove(oZoomList.size() - 1);
					calculate();
					break;
				}
				case RESET: {
					oZoomList.clear();
					calculate();
					break;
				}
				case RESIZE: {
					// this.oTextureRenderer.
					this.renderer.clearBuffers();
					this.rootNode.clearControllers();
					this.rootNode.detachAllChildren();
					this.oTextureRenderer.cleanup();
					oState.deleteAll();
					oRenderState.deleteAll();
					oRenderPrecisionState.deleteAll();
					oRenderInitState.deleteAll();
					oRenderState2.deleteAll();
					oRenderPrecisionState2.deleteAll();
					this.renderer.cleanup();
					// DisplaySystem.getDisplaySystem().reset();
					this.simpleSetup();
					break;
				}

				}
				state = NOTHING;
			}

			Thread.sleep(30);
		} catch (Exception ex) {
			String exString = ex.toString() + " at:\n";
			StackTraceElement[] elements = ex.getStackTrace();
			for (int i = 0; i < elements.length; i++, exString += "\n") {
				exString += elements[i].toString();
			}
			this.oMainFrame.stopIt(exString);
		}
	}

	public void resizeCanvas(int width, int height) {
		super.resizeCanvas(width, height);
		if (this.width != width || this.height != height) {
			this.width = width;
			this.height = height;
			this.state = RESIZE;
			this.oFractalConfigJPanel.setProgressValue(0);
		}
	}
}
